import fs from 'fs'
import { BaseEvent } from '@/application/base/baseEvent'
import { SocketEvent } from '@/application/enums'
import type { UUID } from '@/application/types'
import { SpriteAction } from '@/entities/spriteAction'
import SpriteRepository from '@/repositories/spriteRepository'
import sharp from 'sharp'

interface SpriteImage {
  url: string
  offset: {
    x: number
    y: number
  }
}

interface SpriteActionData {
  action: string
  sprites: SpriteImage[]
  originX: number
  originY: number
  frameRate: number
}

interface UpdateSpritePayload {
  id: UUID
  name: string
  spriteActions: SpriteActionData[]
}

export default class SpriteUpdateEvent extends BaseEvent {
  private readonly spriteRepository: SpriteRepository = new SpriteRepository()

  public listen(): void {
    this.socket.on(SocketEvent.GM_SPRITE_UPDATE, this.handleEvent.bind(this))
  }

  private async handleEvent(data: UpdateSpritePayload, callback: (success: boolean) => void): Promise<void> {
    try {
      // Validate request and permissions
      if (!(await this.isCharacterGM())) {
        callback(false)
        return
      }

      // Get and validate sprite
      const sprite = await this.spriteRepository.getById(data.id)
      if (!sprite) {
        callback(false)
        return
      }
      await this.spriteRepository.getEntityManager().populate(sprite, ['spriteActions'])

      // Update sprite name
      await sprite.setName(data.name).setUpdatedAt(new Date()).save()

      // Process all sprite actions
      const success = await this.processAllSpriteActions(data.spriteActions, sprite)
      callback(success)
    } catch (error) {
      console.error(`Error updating sprite ${data.id}:`, error)
      callback(false)
    }
  }

  private async processAllSpriteActions(actionsData: SpriteActionData[], sprite: any): Promise<boolean> {
    try {
      // Remove existing actions
      const existingActions = sprite.getSpriteActions()
      for (const existingAction of existingActions) {
        await this.spriteRepository.getEntityManager().removeAndFlush(existingAction)
      }

      // First pass: find the global maximum dimensions across all actions
      let globalMaxWidth = 0
      let globalMaxHeight = 0

      // Extract all image metadata to find global maximums
      for (const actionData of actionsData) {
        if (actionData.sprites.length === 0) continue

        const imagesData = await Promise.all(
          actionData.sprites.map(async (sprite) => {
            const base64Data = sprite.url.split(';base64,').pop()
            if (!base64Data) throw new Error('Invalid base64 image')
            const buffer = Buffer.from(base64Data, 'base64')
            const metadata = await sharp(buffer).metadata()

            return {
              width: metadata.width || 0,
              height: metadata.height || 0
            }
          })
        )

        // Update global maximums with this action's maximums
        const actionMaxWidth = Math.max(...imagesData.map((data) => data.width), 0)
        const actionMaxHeight = Math.max(...imagesData.map((data) => data.height), 0)

        globalMaxWidth = Math.max(globalMaxWidth, actionMaxWidth)
        globalMaxHeight = Math.max(globalMaxHeight, actionMaxHeight)
      }

      // Process each action using the global maximum dimensions
      for (const actionData of actionsData) {
        if (actionData.sprites.length === 0) continue

        // Generate and save the sprite sheet using global dimensions
        const frameDimensions = await this.generateAndSaveSpriteSheet(actionData.sprites, sprite.getId(), actionData.action, globalMaxWidth, globalMaxHeight)

        if (!frameDimensions) return false

        // Create and save sprite action
        const spriteAction = new SpriteAction()
        spriteAction.setSprite(sprite)
        sprite.getSpriteActions().add(spriteAction)

        spriteAction
          .setAction(actionData.action)
          .setSprites(actionData.sprites)
          .setOriginX(actionData.originX)
          .setOriginY(actionData.originY)
          .setFrameWidth(frameDimensions.frameWidth)
          .setFrameHeight(frameDimensions.frameHeight)
          .setFrameRate(actionData.frameRate)

        await this.spriteRepository.getEntityManager().persistAndFlush(spriteAction)
      }

      return true
    } catch (error) {
      console.error('Error processing sprite actions:', error)
      return false
    }
  }

  private async generateAndSaveSpriteSheet(sprites: SpriteImage[], spriteId: string, action: string, maxWidth: number, maxHeight: number): Promise<{ frameWidth: number; frameHeight: number } | null> {
    try {
      if (sprites.length === 0) return { frameWidth: 0, frameHeight: 0 }

      // Extract image data
      const imagesData = await Promise.all(
        sprites.map(async (sprite) => {
          const base64Data = sprite.url.split(';base64,').pop()
          if (!base64Data) throw new Error('Invalid base64 image')
          const buffer = Buffer.from(base64Data, 'base64')
          const metadata = await sharp(buffer).metadata()

          return {
            buffer,
            width: metadata.width || 0,
            height: metadata.height || 0
          }
        })
      )

      // Skip creation if any image has invalid dimensions
      if (imagesData.some((data) => data.width === 0 || data.height === 0)) {
        console.error('One or more sprites have invalid dimensions')
        return null
      }

      // Create frames of uniform size with the original sprites centered
      const uniformFrames = await Promise.all(
        imagesData.map(async (imageData) => {
          // Calculate centering offsets to position the sprite in the middle of the frame
          const xOffset = Math.floor((maxWidth - imageData.width) / 2)
          const yOffset = Math.floor((maxHeight - imageData.height) / 2)

          // Create a uniform-sized frame with the sprite centered
          return sharp({
            create: {
              width: maxWidth,
              height: maxHeight,
              channels: 4,
              background: { r: 0, g: 0, b: 0, alpha: 0 }
            }
          })
            .composite([
              {
                input: imageData.buffer,
                left: xOffset,
                top: yOffset
              }
            ])
            .png()
            .toBuffer()
        })
      )

      // Create the sprite sheet with uniform frames
      const spriteSheet = await sharp({
        create: {
          width: maxWidth * uniformFrames.length,
          height: maxHeight,
          channels: 4,
          background: { r: 0, g: 0, b: 0, alpha: 0 }
        }
      })
        .composite(
          uniformFrames.map((buffer, index) => ({
            input: buffer,
            left: index * maxWidth,
            top: 0
          }))
        )
        .png()
        .toBuffer()

      // Save sprite sheet
      const dir = `public/sprites/${spriteId}`
      await fs.promises.mkdir(dir, { recursive: true })
      await fs.promises.writeFile(`${dir}/${action}.png`, spriteSheet)

      // Return the uniform frame dimensions (now global maximum dimensions)
      return {
        frameWidth: maxWidth,
        frameHeight: maxHeight
      }
    } catch (error) {
      console.error('Error generating sprite sheet:', error)
      return null
    }
  }
}
